.gp-image-menu-item > image {
  min-height: 16px;
  min-width: 16px;
}

.gp-image-menu-item:not(.image-only) > image:dir(ltr) {
  margin-right: 5px;
}

.gp-image-menu-item:not(.image-only) > image:dir(rtl) {
  margin-left: 5px;
}

panel-toplevel.left gp-applet > menubar > .gp-image-menu-item:not(.image-only) > image:dir(ltr) {
  margin-right: 0px;
  margin-top: 5px;
}

panel-toplevel.left gp-applet > menubar > .gp-image-menu-item:not(.image-only) > image:dir(rtl) {
  margin-left: 0px;
  margin-bottom: 5px;
}

panel-toplevel.right gp-applet > menubar > .gp-image-menu-item:not(.image-only) > image:dir(ltr) {
  margin-right: 0px;
  margin-bottom: 5px;
}

panel-toplevel.right gp-applet > menubar > .gp-image-menu-item:not(.image-only) > image:dir(rtl) {
  margin-left: 0px;
  margin-top: 5px;
}

gp-menu-button:hover .icon,
gp-action-button:hover image,
gp-launcher-button:hover image {
  -gtk-icon-effect: highlight;
}

panel-toplevel.horizontal gp-menu-button,
panel-toplevel.horizontal gp-action-button,
panel-toplevel.horizontal gp-launcher-button  {
  padding: 0 2px;
}

panel-toplevel.vertical gp-menu-button,
panel-toplevel.vertical gp-action-button,
panel-toplevel.vertical gp-launcher-button {
  padding: 2px 0;
}

panel-toplevel.horizontal gp-menu-button .icon {
  padding: 0 4px;
}

panel-toplevel.vertical gp-menu-button .icon {
  padding: 4px 0;
}

na-tray-applet {
  -NaApplet-icon-padding: 4;
}

gp-applet.separator > separator {
  margin: 2px 4px;
}

.horizontal sn-item,
.horizontal .system-indicators > menubar > menuitem {
  padding: 0px 4px;
}

.vertical sn-item,
.vertical .system-indicators > menubar > menuitem {
  padding: 4px 0px;
}

#tasklist-button,
#clock-applet-button,
#showdesktop-button {
  padding: 0px 4px;
  border: 0;
  border-radius: 0;
}

gp-arrow-button {
  min-width: 20px;
  min-height: 0px;
  border-radius: 3px;
  margin: 2px;
}

.vertical gp-arrow-button {
  min-width: 0px;
  min-height: 20px;
}

gp-handle {
  -gtk-icon-source: -gtk-recolor(url("/org/gnome/gnome-panel/theme/gp-handle-symbolic.svg"));
  min-width: 16px;
  min-height: 16px;
}

.horizontal gp-handle {
  margin: 0 -3px;
}

.vertical gp-handle {
  -gtk-icon-transform: rotate(90deg);
  margin: -3px 0;
}

.context-row {
  padding: 4px 10px;
}

.calendar-window-button {
  border-radius: 3px;
  padding: 4px 8px;
  min-height: 16px;
}

.calendar-window-button label {
  font-size: smaller;
}

gp-applet.error button {
  border-radius: 3px;
}

.horizontal gp-applet.error button {
  padding: 0 4px;
}

.vertical gp-applet.error button {
  padding: 4px 0;
}
