#ifndef PANEL_ICON_NAMES_H
#define PANEL_ICON_NAMES_H

#define PANEL_ICON_BURNER		"nautilus-cd-burner"
#define PANEL_ICON_COMPUTER		"computer"
#define PANEL_ICON_DESKTOP		"user-desktop"
#define PANEL_ICON_FILESYSTEM		"drive-harddisk"
#define PANEL_ICON_FOLDER		"folder"
#define PANEL_ICON_FOLDER_DOCUMENTS	"folder-documents"
#define PANEL_ICON_FOLDER_DOWNLOAD	"folder-download"
#define PANEL_ICON_FOLDER_MUSIC		"folder-music"
#define PANEL_ICON_FOLDER_PICTURES	"folder-pictures"
#define PANEL_ICON_FOLDER_PUBLIC_SHARE	"folder-publicshare"
#define PANEL_ICON_FOLDER_TEMPLATES	"folder-templates"
#define PANEL_ICON_FOLDER_VIDEOS	"folder-videos"
#define PANEL_ICON_HOME			"user-home"
#define PANEL_ICON_NETWORK		"network-workgroup"
#define PANEL_ICON_PANEL		"gnome-panel"
#define PANEL_ICON_REMOTE		"applications-internet"
#define PANEL_ICON_UNKNOWN		"image-missing"

#endif /* PANEL_ICON_NAMES_H */
